#include "controlworker.h"
#include "frontend_nodes.h"
#include "frontend_param.h"
#include "frontend_source.h"
#include "graph_blocks.h"
#include <QProcess>
#include <err.h>

ControlWorker::ControlWorker(QObject *parent) :
    QObject(parent), conn(NULL)
{

}

ControlWorker::~ControlWorker()
{
    disconnectBackend();
}

void ControlWorker::connectBackend()
{
    conn = ucart_backendConnect();
}

void ControlWorker::disconnectBackend()
{
    if (conn) {
         ucart_backendDisconnect(conn);
         conn = NULL;
    }
}

void ControlWorker::getNodes()
{
    if (conn) {
        frontend_node_data * nd = NULL;
        size_t num_nodes = 0;
        frontend_getnodes(conn, &nd, &num_nodes);
        QStringList nodes;
        QStringList const_block_nodes;
        for (size_t i = 0; i < num_nodes; i++) {
            nodes.append(QString(nd[i].name));
            if ((nd[i].type == BLOCK_CONSTANT) || (nd[i].type == BLOCK_ADD)) {
                const_block_nodes.append(nd[i].name);
            }
        }
        emit(gotNodes(nodes));
        emit(gotConstantBlocks(const_block_nodes));

        /* Create a graph for rendering.
         *
         * This leaks memory (NOT MY FAULT!) because there isn't a function to free
         * a graph created with create_graph().
         */
        computation_graph * cgraph = create_graph();

        /* Add nodes */
        for (size_t i = 0; i < num_nodes; i++) {
            graph_add_node_id(cgraph, nd[i].block, nd[i].name, blockDefs[nd[i].type]);
        }

        /* Set sources */
        for (size_t i = 0; i < num_nodes; i++) {
            for (ssize_t j = 0; j < blockDefs[nd[i].type]->n_inputs; j++) {
                frontend_source_data sd;
                sd.dst_block = nd[i].block;
                sd.dst_input = j;
                sd.src_block = 0;
                sd.src_output = 0;
                frontend_getsource(conn, &sd);
                graph_set_source(cgraph, sd.dst_block, sd.dst_input, sd.src_block, sd.src_output);
            }
        }

        /* Set params */
        for (size_t i = 0; i < num_nodes; i++) {
            for (ssize_t j = 0; j < blockDefs[nd[i].type]->n_params; j++) {
                frontend_param_data pd;
                pd.block = nd[i].block;
                pd.param = j;
                pd.value = 0;
                frontend_getparam(conn, &pd);
                graph_set_param_val(cgraph, pd.block, pd.param, pd.value);
            }
        }

        /* This is a massive hack... */
        /* Could be done with popen, but fuck it */
        FILE * dotfile = fopen("/tmp/ucart-tmp-graph.dot", "w");
        if (dotfile) {
            export_dot(cgraph, dotfile, 0);
            fclose(dotfile);
            if (int dot_exit = QProcess::execute("dot /tmp/ucart-tmp-graph.dot -Tpng -o /tmp/ucart-tmp-graph.png")) {
                warnx("dot returned nonzero value (%d)", dot_exit);
            } else {
                emit(graphRendered(QString("/tmp/ucart-tmp-graph.png")));
            }
        } else {
            warn("fopen (/tmp/ucart-tmp-graph.dot)");
        }

        /* And here's where I'd put a call to free_graph(), IF I HAD ONE! */
        frontend_free_node_data(nd, num_nodes);
    }
}

void ControlWorker::getParams(QString node)
{
    if (conn) {
        frontend_node_data * nd = NULL;
        size_t num_nodes = 0;
        frontend_getnodes(conn, &nd, &num_nodes);

        for (size_t i = 0; i < num_nodes; i++) {
            if (QString(nd[i].name) == node) {
                QStringList params;

                /* Get type definition */
                const struct graph_node_type * type = blockDefs[nd[i].type];
                /* Iterate through param names to append, and then emit signal */
                for (ssize_t j = 0; j < type->n_params; j++) {
                    params.append(QString(type->param_names[j]));
                }
                emit(gotParams(params));
            }
        }
        frontend_free_node_data(nd, num_nodes);
    }
}



void ControlWorker::getParamValue(QString node, QString param)
{
    if (conn) {
        frontend_node_data * nd = NULL;
        size_t num_nodes = 0;
        frontend_getnodes(conn, &nd, &num_nodes);

        for (size_t i = 0; i < num_nodes; i++) {
            if (QString(nd[i].name) == node) {
                frontend_param_data pd;
                pd.block = nd[i].block;

                /* Get the type definition, then iterate through to find the param */
                const struct graph_node_type * type = blockDefs[nd[i].type];
                for (ssize_t j = 0; j < type->n_params; j++) {
                    /* Found param */
                    if (QString(type->param_names[j]) == param) {
                        /* Set pd.param and finish the job */
                        pd.param = j;
                        frontend_getparam(conn, &pd);
                        emit(gotParamValue(node, param, pd.value));
                    }
                }
            }
        }
        frontend_free_node_data(nd, num_nodes);
    }
}

void ControlWorker::setParamValue(QString node, QString param, float value)
{
    if (conn) {
        frontend_node_data * nd = NULL;
        size_t num_nodes = 0;
        frontend_getnodes(conn, &nd, &num_nodes);

        for (size_t i = 0; i < num_nodes; i++) {
            if (QString(nd[i].name) == node) {
                frontend_param_data pd;
                pd.block = nd[i].block;

                const struct graph_node_type * type = blockDefs[nd[i].type];
                for (ssize_t j = 0; j < type->n_params; j++) {
                    if (QString(type->param_names[j]) == param) {
                        pd.param = j;
                        pd.value = value;
                        frontend_setparam(conn, &pd);
                        emit(paramSet(node, param));
                    }
                }
            }
        }
        frontend_free_node_data(nd, num_nodes);
    }
}
